@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2021 Hannes Domani
+ * Copyright (C) 2013-2025 Hannes Domani
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -25,20 +25,26 @@
 #include <string.h>
 
 
-typedef void ChildWalker( Dwarf_Debug dbg,Dwarf_Die die,void *context );
+typedef int ChildWalker( Dwarf_Debug dbg,Dwarf_Die die,void *context );
 
 // call walkFunc() recursively for every child DIE
-static void walkChilds( Dwarf_Debug dbg,Dwarf_Die die,
+static int walkChilds( Dwarf_Debug dbg,Dwarf_Die die,
     ChildWalker *walkFunc,void *context )
 {
   Dwarf_Die child;
   if( dwarf_child(die,&child,NULL)!=DW_DLV_OK )
-    return;
+    return( 0 );
 
   while( 1 )
   {
-    walkChilds( dbg,child,walkFunc,context );
-    walkFunc( dbg,child,context );
+    int stopChilds = walkChilds( dbg,child,walkFunc,context );
+    int stopCur = walkFunc( dbg,child,context );
+
+    if( stopChilds || stopCur )
+    {
+      dwarf_dealloc( dbg,child,DW_DLA_DIE );
+      return( 1 );
+    }
 
     Dwarf_Die next_child;
     int res = dwarf_siblingof_b( dbg,child,1,&next_child,NULL );
@@ -48,6 +54,8 @@ static void walkChilds( Dwarf_Debug dbg,Dwarf_Die die,
     if( res!=DW_DLV_OK ) break;
     child = next_child;
   }
+
+  return( 0 );
 }
 
 
@@ -262,19 +270,19 @@ typedef struct inline_info
 } inline_info;
 
 // find the calling-location of inlined functions
-static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
+static int findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
 {
   Dwarf_Half tag;
   if( dwarf_tag(die,&tag,NULL)!=DW_DLV_OK ||
       (tag!=DW_TAG_inlined_subroutine &&
        tag!=DW_TAG_subprogram) )
-    return;
+    return( 0 );
 
   Dwarf_Addr low,high;
   if( dwarf_lowhighpc(die,&low,&high)==DW_DLV_OK && high )
   {
     if( cuInfo->ptr<low || cuInfo->ptr>=high )
-      return;
+      return( 0 );
   }
   else
   {
@@ -285,7 +293,7 @@ static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
     Dwarf_Unsigned rngEntriesCount;
     if( dwarf_ranges(dbg,die,&version,&ranges,&rangeCount,
           &rnghlhead,&rngEntriesCount)!=DW_DLV_OK )
-      return;
+      return( 0 );
 
     if( version<=4 )
     {
@@ -318,7 +326,7 @@ static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
 
       dwarf_dealloc_ranges( dbg,ranges,rangeCount );
 
-      if( i>=rangeCount ) return;
+      if( i>=rangeCount ) return( 0 );
     }
     else
     {
@@ -327,13 +335,11 @@ static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
       {
         unsigned entrylen = 0;
         unsigned code = 0;
-        Dwarf_Unsigned rawlowpc = 0;
-        Dwarf_Unsigned rawhighpc = 0;
         Dwarf_Unsigned lowpc = 0;
         Dwarf_Unsigned highpc = 0;
         Dwarf_Bool debug_addr_unavailable = 0;
         int res = dwarf_get_rnglists_entry_fields_a( rnghlhead,i,
-            &entrylen,&code,&rawlowpc,&rawhighpc,
+            &entrylen,&code,NULL,NULL,
             &debug_addr_unavailable,&lowpc,&highpc,NULL );
         if( res!=DW_DLV_OK || code==DW_RLE_end_of_list )
         {
@@ -352,7 +358,7 @@ static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
 
       dwarf_dealloc_rnglists_head( rnghlhead );
 
-      if( i>=rngEntriesCount ) return;
+      if( i>=rngEntriesCount ) return( 0 );
     }
   }
 
@@ -366,18 +372,18 @@ static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
 
     if( funcname )
       dwarf_dealloc( dbg,funcname,DW_DLA_STRING );
-    return;
+    return( 1 );
   }
 
   Dwarf_Attribute callfile;
   if( dwarf_attr(die,DW_AT_call_file,&callfile,NULL)!=DW_DLV_OK )
-    return;
+    return( 1 );
 
   Dwarf_Attribute callline;
   if( dwarf_attr(die,DW_AT_call_line,&callline,NULL)!=DW_DLV_OK )
   {
     dwarf_dealloc( dbg,callfile,DW_DLA_ATTR );
-    return;
+    return( 1 );
   }
 
   Dwarf_Unsigned fileno,lineno;
@@ -413,8 +419,15 @@ static void findInlined( Dwarf_Debug dbg,Dwarf_Die die,inline_info *cuInfo )
 
   dwarf_dealloc( dbg,callfile,DW_DLA_ATTR );
   dwarf_dealloc( dbg,callline,DW_DLA_ATTR );
+
+  return( 1 );
 }
 
+typedef struct range_t
+{
+  Dwarf_Addr low;
+  Dwarf_Addr high;
+} range_t;
 
 typedef struct cu_info
 {
@@ -426,6 +439,8 @@ typedef struct cu_info
   int fileno_offs;
   char **files;
   Dwarf_Signed fileCount;
+  range_t *ranges;
+  int rangeCount;
 } cu_info;
 
 int dwstOfFileExt(
@@ -455,7 +470,6 @@ int dwstOfFileExt(
 
   cu_info *cuArr = NULL;
   int cuQty = 0;
-  Dwarf_Addr lowestLow = 0;
   while( 1 )
   {
     Dwarf_Unsigned next_cu_header;
@@ -483,13 +497,15 @@ int dwstOfFileExt(
     if( dwarf_dieoffset(die,&cuInfo->offs,NULL)!=DW_DLV_OK )
       cuInfo->offs = 0;
 
+    cuInfo->ranges = NULL;
+    cuInfo->rangeCount = 0;
+
+    cuInfo->low = 0;
+    cuInfo->high = 0;
     int res = dwarf_lowhighpc( die,&cuInfo->low,&cuInfo->high );
-    if( res==DW_DLV_OK && cuInfo->low &&
-        (!lowestLow || cuInfo->low<lowestLow) )
-      lowestLow = cuInfo->low;
     if( res!=DW_DLV_OK || !cuInfo->high )
     {
-      int hasLow = res==DW_DLV_OK;
+      int hasLow = res==DW_DLV_OK && cuInfo->low;
       if( !hasLow ) cuInfo->low = 0;
       cuInfo->high = 0;
 
@@ -504,7 +520,9 @@ int dwstOfFileExt(
         if( version<=4 )
         {
           int i;
-          Dwarf_Addr base = 0;
+          Dwarf_Addr base = cuInfo->low;
+          cuInfo->ranges = malloc( rangeCount*sizeof(range_t) );
+          if( !cuInfo->ranges ) rangeCount = 0;
           for( i=0; i<rangeCount; i++ )
           {
             Dwarf_Ranges *range = ranges + i;
@@ -533,6 +551,10 @@ int dwstOfFileExt(
             }
             if( high>cuInfo->high )
               cuInfo->high = high;
+
+            cuInfo->ranges[cuInfo->rangeCount].low = low;
+            cuInfo->ranges[cuInfo->rangeCount].high = high;
+            cuInfo->rangeCount++;
           }
 
           dwarf_dealloc_ranges( dbg,ranges,rangeCount );
@@ -540,6 +562,8 @@ int dwstOfFileExt(
         else
         {
           unsigned i;
+          cuInfo->ranges = malloc( rngEntriesCount*sizeof(range_t) );
+          if( !cuInfo->ranges ) rngEntriesCount = 0;
           for( i=0; i<rngEntriesCount; i++ )
           {
             unsigned entrylen = 0;
@@ -558,7 +582,7 @@ int dwstOfFileExt(
               break;
             }
             if( code==DW_RLE_base_addressx || code==DW_RLE_base_address ||
-                debug_addr_unavailable )
+                debug_addr_unavailable || !lowpc )
               continue;
 
             if( !hasLow || lowpc<cuInfo->low )
@@ -568,6 +592,10 @@ int dwstOfFileExt(
             }
             if( highpc>cuInfo->high )
               cuInfo->high = highpc;
+
+            cuInfo->ranges[cuInfo->rangeCount].low = lowpc;
+            cuInfo->ranges[cuInfo->rangeCount].high = highpc;
+            cuInfo->rangeCount++;
           }
 
           dwarf_dealloc_rnglists_head( rnghlhead );
@@ -592,7 +620,7 @@ int dwstOfFileExt(
       baseOffs = imageBase_dbg - imageBase;
   }
 
-  int i,j;
+  int i,j,k;
   for( i=0; i<count; i++ )
   {
     uint64_t ptrOrig = addr[i];
@@ -602,7 +630,13 @@ int dwstOfFileExt(
     for( j=0; j<cuQty; j++ )
     {
       cu_info *cuInfo = &cuArr[j];
-      if( ptr<cuInfo->low || ptr>=cuInfo->high ) continue;
+      if( cuInfo->high && (ptr<cuInfo->low || ptr>=cuInfo->high) )
+        continue;
+      for( k=0; k<cuInfo->rangeCount; ++k )
+        if( ptr>=cuInfo->ranges[k].low && ptr<cuInfo->ranges[k].high )
+          break;
+      if( k && k>=cuInfo->rangeCount )
+        continue;
 
       Dwarf_Die die;
       if( cuInfo->offs &&
@@ -720,6 +754,8 @@ int dwstOfFileExt(
 
       dwarf_dealloc( dbg,files,DW_DLA_LIST );
     }
+
+    free( cuArr[j].ranges );
   }
   free( cuArr );
 
